/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <unistd.h>
#include <malloc.h>
#include <arpa/inet.h>
#include "mlx5dv_dr.h"
#include "mlx5.h"
#include "mlx5_ifc.h"
#include <arpa/inet.h>
#include <ccan/array_size.h>

/* helper macros */
#define __dr_action_nullp(typ) ((struct typ##_bits *)0)
#define __dr_action_bit_sz(typ, fld) sizeof(__dr_action_nullp(typ)->fld)
#define __dr_action_bit_off(typ, fld) (offsetof(struct typ##_bits, fld))
#define __dr_action_dw_off(typ, fld) (__dr_action_bit_off(typ, fld) / 32)
#define __dr_action_dw_bit_off(typ, fld) (32 - __dr_action_bit_sz(typ, fld) - (__dr_action_bit_off(typ, fld) & 0x1f))
#define __dr_action_mask(typ, fld) ((uint32_t)((1ull << __dr_action_bit_sz(typ, fld)) - 1))
#define __dr_action_dw_mask(typ, fld) (__dr_action_mask(typ, fld) << __dr_action_dw_bit_off(typ, fld))
#define __dr_action_st_sz_bits(typ) sizeof(struct typ##_bits)

#define DR_ACTION_ST_SZ_BYTES(typ) (sizeof(struct typ##_bits) / 8)
#define DR_ACTION_ST_SZ_DW(typ) (sizeof(struct typ##_bits) / 32)
#define DR_ACTION_ST_SZ_QW(typ) (sizeof(struct typ##_bits) / 64)

/* Write to layout a value */
#define DR_ACTION_SET(typ, p, fld, v) do { \
	uint32_t _v = v; \
	static_assert(!(__dr_action_st_sz_bits(typ) % 32), \
		      "struct size must be a multiple of 32 bit dwords"); \
	*((__be32 *)(p) + __dr_action_dw_off(typ, fld)) = \
	htonl((ntohl(*((__be32 *)(p) + __dr_action_dw_off(typ, fld))) & \
		     (~__dr_action_dw_mask(typ, fld))) | (((_v) & __dr_action_mask(typ, fld)) \
		     << __dr_action_dw_bit_off(typ, fld))); \
} while (0)

/* Read from layout struct */
#define DR_ACTION_GET(typ, p, fld) \
	((ntohl(*((__be32 *)(p) + \
	__dr_action_dw_off(typ, fld))) >> __dr_action_dw_bit_off(typ, fld)) & \
	__dr_action_mask(typ, fld))

enum dr_action_domain {
	DR_ACTION_DOMAIN_NIC_INGRESS,
	DR_ACTION_DOMAIN_NIC_EGRESS,
	DR_ACTION_DOMAIN_FDB_INGRESS,
	DR_ACTION_DOMAIN_FDB_EGRESS,
	DR_ACTION_DOMAIN_RESERVED,
};

enum dr_action_valid_state {
	DR_ACTION_STATE_ERR,
	DR_ACTION_STATE_NO_ACTION,
	DR_ACTION_STATE_REFORMAT,
	DR_ACTION_STATE_MODIFY_HDR,
	DR_ACTION_STATE_MODIFY_VLAN,
	DR_ACTION_STATE_NON_TERM,
	DR_ACTION_STATE_TERM,
	DR_ACTION_STATE_RESERVED,
};

static enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_RESERVED]
						   [DR_ACTION_STATE_RESERVED]
						   [DR_ACTION_TYP_RESERVED] = {
	[DR_ACTION_DOMAIN_NIC_INGRESS] = {
		[DR_ACTION_STATE_NO_ACTION] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TAG]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TNL_L2_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_TNL_L3_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_REFORMAT] = {
			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TAG]		= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_MODIFY_HDR] = {
			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TAG]		= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_NON_TERM] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TAG]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TNL_L2_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_TNL_L3_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_TERM] = {
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
		},
	},
	[DR_ACTION_DOMAIN_NIC_EGRESS] = {
		[DR_ACTION_STATE_NO_ACTION] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_REFORMAT] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_MODIFY_HDR] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
		},
		[DR_ACTION_STATE_NON_TERM] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_TERM] = {
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
		},
	},
	// TODO: In the future we might be able to combine RX and FDB_RX
	[DR_ACTION_DOMAIN_FDB_INGRESS] = {
		[DR_ACTION_STATE_NO_ACTION] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TNL_L2_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_TNL_L3_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_POP_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_REFORMAT] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_POP_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_MODIFY_HDR] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_MODIFY_VLAN] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_POP_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
		},
		[DR_ACTION_STATE_NON_TERM] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_TNL_L2_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_TNL_L3_TO_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_POP_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_TERM] = {
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
		},
	},
	// TODO: In the future we might be able to combine TX and FDB_TX
	[DR_ACTION_DOMAIN_FDB_EGRESS] = {
		[DR_ACTION_STATE_NO_ACTION] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_REFORMAT] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_MODIFY_HDR] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_MODIFY_VLAN] = {
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_NON_TERM] = {
			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_NON_TERM,
			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_REFORMAT,
			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_MODIFY_VLAN,
			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
		},
		[DR_ACTION_STATE_TERM] = {
			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
		},
	},
};

enum {
	STE_VLAN_NOT_PRESENT,
	STE_SVLAN,
	STE_CVLAN
};

/* Both HW set and HW add share the same HW format with different opcodes */
struct dr_action_struct_hw_set_bits {
	uint8_t opcode[0x8];
	uint8_t destination_field_code[0x8];
	uint8_t reserved_at_10[0x2];
	uint8_t destination_left_shifter[0x6];
	uint8_t reserved_at_18[0x3];
	uint8_t destination_length[0x5];
	uint8_t inline_data[0x20];
};

enum {
	DR_ACTION_MDFY_HW_FLD_L2_0	= 0,
	DR_ACTION_MDFY_HW_FLD_L2_1	= 1,
	DR_ACTION_MDFY_HW_FLD_L2_2	= 2,
	DR_ACTION_MDFY_HW_FLD_L3_0	= 3,
	DR_ACTION_MDFY_HW_FLD_L3_1	= 4,
	DR_ACTION_MDFY_HW_FLD_L3_2	= 5,
	DR_ACTION_MDFY_HW_FLD_L3_3	= 6,
	DR_ACTION_MDFY_HW_FLD_L3_4	= 7,
	DR_ACTION_MDFY_HW_FLD_L4_0	= 8,
	DR_ACTION_MDFY_HW_FLD_L4_1	= 9,
	DR_ACTION_MDFY_HW_FLD_MPLS	= 10,
	DR_ACTION_MDFY_HW_FLD_L2_TNL_0	= 11,
	DR_ACTION_MDFY_HW_FLD_REG_0	= 12,
	DR_ACTION_MDFY_HW_FLD_REG_1	= 13,
	DR_ACTION_MDFY_HW_FLD_REG_2	= 14,
	DR_ACTION_MDFY_HW_FLD_REG_3	= 15,
	DR_ACTION_MDFY_HW_FLD_L4_2	= 16,
	DR_ACTION_MDFY_HW_FLD_FLEX_0	= 17,
	DR_ACTION_MDFY_HW_FLD_FLEX_1	= 18,
	DR_ACTION_MDFY_HW_FLD_FLEX_2	= 19,
	DR_ACTION_MDFY_HW_FLD_FLEX_3	= 20,
	DR_ACTION_MDFY_HW_FLD_L2_TNL_1	= 21,
	DR_ACTION_MDFY_HW_FLD_METADATA	= 22,
	DR_ACTION_MDFY_HW_FLD_RESERVED	= 23,
};

enum dr_action_modify_hw_op {
	DR_ACTION_MDFY_HW_OP_SET	= 0x2,
	DR_ACTION_MDFY_HW_OP_ADD	= 0x3,
};

enum dr_action_modify_hw_l3_hdr {
	DR_ACTION_MDFY_HW_HDR_L3_NONE	= 0x0,
	DR_ACTION_MDFY_HW_HDR_L3_IPV4	= 0x1,
	DR_ACTION_MDFY_HW_HDR_L3_IPV6	= 0x2,
};

enum dr_action_modify_hw_l4_hdr {
	DR_ACTION_MDFY_HW_HDR_L4_NONE	= 0x0,
	DR_ACTION_MDFY_HW_HDR_L4_TCP	= 0x1,
	DR_ACTION_MDFY_HW_HDR_L4_UDP	= 0x2,
};

struct dr_action_modify_field_conv {
	uint16_t hw_field;
	uint8_t start;
	uint8_t end;
	uint8_t l3_type;
	uint8_t l4_type;
};

struct dr_action_modify_field_conv dr_action_conv_arr[] = {
	[MLX5_ACTION_IN_FIELD_OUT_SMAC_47_16] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L2_1, .start = 16, .end = 47,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SMAC_15_0] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L2_1, .start = 0, .end = 15,
	},
	[MLX5_ACTION_IN_FIELD_OUT_ETHERTYPE] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L2_2, .start = 32, .end = 47,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DMAC_47_16] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L2_0, .start = 16, .end = 47,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DMAC_15_0] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L2_0, .start = 0, .end = 15,
	},
	[MLX5_ACTION_IN_FIELD_OUT_IP_DSCP] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_1, .start = 0, .end = 5,
	},
	[MLX5_ACTION_IN_FIELD_OUT_TCP_FLAGS] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_0, .start = 48, .end = 56,
		.l4_type = DR_ACTION_MDFY_HW_HDR_L4_TCP,
	},
	[MLX5_ACTION_IN_FIELD_OUT_TCP_SPORT] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_0, .start = 0, .end = 15,
		.l4_type = DR_ACTION_MDFY_HW_HDR_L4_TCP,
	},
	[MLX5_ACTION_IN_FIELD_OUT_TCP_DPORT] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_0, .start = 16, .end = 31,
		.l4_type = DR_ACTION_MDFY_HW_HDR_L4_TCP,
	},
	[MLX5_ACTION_IN_FIELD_OUT_IP_TTL] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_1, .start = 8, .end = 15,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV4,
	},
	[MLX5_ACTION_IN_FIELD_OUT_IPV6_HOPLIMIT] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_1, .start = 8, .end = 15,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_UDP_SPORT] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_0, .start = 0, .end = 15,
		.l4_type = DR_ACTION_MDFY_HW_HDR_L4_UDP,
	},
	[MLX5_ACTION_IN_FIELD_OUT_UDP_DPORT] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_0, .start = 16, .end = 31,
		.l4_type = DR_ACTION_MDFY_HW_HDR_L4_UDP,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SIPV6_127_96] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_3, .start = 32, .end = 63,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SIPV6_95_64] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_3, .start = 0, .end = 31,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SIPV6_63_32] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_4, .start = 32, .end = 63,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SIPV6_31_0] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_4, .start = 0, .end = 31,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DIPV6_127_96] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_0, .start = 32, .end = 63,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DIPV6_95_64] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_0, .start = 0, .end = 31,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DIPV6_63_32] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_2, .start = 32, .end = 63,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DIPV6_31_0] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_2, .start = 0, .end = 31,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV6,
	},
	[MLX5_ACTION_IN_FIELD_OUT_SIPV4] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_0, .start = 0, .end = 31,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV4,
	},
	[MLX5_ACTION_IN_FIELD_OUT_DIPV4] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L3_0, .start = 32, .end = 63,
		.l3_type = DR_ACTION_MDFY_HW_HDR_L3_IPV4,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGA] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_METADATA, .start = 0, .end = 31,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGB] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_METADATA, .start = 32, .end = 63,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_0] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_0, .start = 32, .end = 63,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_1] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_0, .start = 0, .end = 31,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_2] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_1, .start = 32, .end = 63,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_3] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_1, .start = 0, .end = 31,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_4] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_2, .start = 32, .end = 63,
	},
	[MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGC_5] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_REG_2, .start = 0, .end = 31,
	},
	[MLX5_ACTION_IN_FIELD_OUT_TCP_SEQ_NUM] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_1, .start = 32, .end = 63,
	},
	[MLX5_ACTION_IN_FIELD_OUT_TCP_ACK_NUM] = {
		.hw_field = DR_ACTION_MDFY_HW_FLD_L4_1, .start = 0, .end = 31,
	},
};

#define DR_ACTION_INIT_NEXT_STE(_lu_type) do { \
	last_ste += DR_STE_SIZE; \
	(*added_stes)++; \
	dr_ste_init(last_ste, DR_STE_LU_TYPE_DONT_CARE, _lu_type, gvmi); \
} while (0)

#define MAX_VLANS 2
struct dr_action_vlan_info {
	int count;
	uint32_t headers[MAX_VLANS];
};

static void dr_actions_apply_tx(uint8_t *action_type_set,
				uint8_t *last_ste,
				uint16_t gvmi,
				uint32_t ctr_id,
				uint32_t modify_index,
				uint16_t modify_actions,
				uint32_t reformat_id,
				uint32_t reformat_size,
				struct dr_action_vlan_info *vlans,
				uint32_t *added_stes)
{
	/* We want to make sure the modify header comes before L2
	 * encapsulation. The reason for that is that we support
	 * modify headers for outer headers only
	 */
	if (action_type_set[DR_ACTION_TYP_MODIFY_HDR]) {
		dr_ste_set_entry_type(last_ste, DR_STE_TYPE_MODIFY_PKT);
		dr_ste_set_rewrite_actions(last_ste, modify_actions,
					   modify_index);
	}

	if (action_type_set[DR_ACTION_TYP_PUSH_VLAN]) {
		int i;

		for (i = 0; i < vlans->count; i++) {
			if (i || action_type_set[DR_ACTION_TYP_MODIFY_HDR])
				DR_ACTION_INIT_NEXT_STE(DR_STE_TYPE_TX);

			dr_ste_set_tx_push_vlan(last_ste, vlans->headers[i]);
		}
	}

	if (action_type_set[DR_ACTION_TYP_L2_TO_TNL_L2] ||
	    action_type_set[DR_ACTION_TYP_L2_TO_TNL_L3]) {
		/* - Modify header and encapsulation require a different STE */
		if (action_type_set[DR_ACTION_TYP_MODIFY_HDR] ||
		    action_type_set[DR_ACTION_TYP_PUSH_VLAN])
			DR_ACTION_INIT_NEXT_STE(DR_STE_TYPE_TX);

		dr_ste_set_tx_encap(last_ste, reformat_id, reformat_size,
				    action_type_set[DR_ACTION_TYP_L2_TO_TNL_L3]);
	}

	if (action_type_set[DR_ACTION_TYP_CTR]) {
		dr_ste_set_counter_id(last_ste, ctr_id);
	}
}

#define WITH_VLAN_NUM_HW_ACTIONS 6

static void dr_actions_apply_rx(uint8_t *action_type_set,
				uint8_t *last_ste,
				uint16_t gvmi,
				uint32_t flow_tag,
				uint32_t ctr_id,
				uint32_t modify_index,
				uint16_t modify_actions,
				uint32_t decap_index,
				uint16_t decap_actions,
				struct dr_action_vlan_info *vlans,
				uint32_t *added_stes)
{
	if (action_type_set[DR_ACTION_TYP_CTR])
		dr_ste_set_counter_id(last_ste, ctr_id);

	if (action_type_set[DR_ACTION_TYP_TNL_L3_TO_L2]) {
		dr_ste_set_entry_type(last_ste, DR_STE_TYPE_MODIFY_PKT);
		dr_ste_set_rewrite_actions(last_ste, decap_actions, decap_index);
		dr_ste_set_rx_decap_l3(last_ste, !!(decap_actions == WITH_VLAN_NUM_HW_ACTIONS));
	}

	if (action_type_set[DR_ACTION_TYP_TNL_L2_TO_L2])
		dr_ste_set_rx_decap(last_ste);

	if (action_type_set[DR_ACTION_TYP_POP_VLAN]) {
		int i;

		for (i = 0; i < vlans->count; i++) {
			if (i ||
			    action_type_set[DR_ACTION_TYP_TNL_L2_TO_L2] ||
			    action_type_set[DR_ACTION_TYP_TNL_L3_TO_L2])
				DR_ACTION_INIT_NEXT_STE(DR_STE_TYPE_RX);

			dr_ste_set_rx_pop_vlan(last_ste);
		}
	}

	if (action_type_set[DR_ACTION_TYP_MODIFY_HDR]) {
		if (dr_ste_get_entry_type(last_ste) == DR_STE_TYPE_MODIFY_PKT)
			DR_ACTION_INIT_NEXT_STE(DR_STE_TYPE_MODIFY_PKT);
		else
			dr_ste_set_entry_type(last_ste, DR_STE_TYPE_MODIFY_PKT);

		dr_ste_set_rewrite_actions(last_ste, modify_actions, modify_index);
	}

	if (action_type_set[DR_ACTION_TYP_TAG]) {
		if (dr_ste_get_entry_type(last_ste) == DR_STE_TYPE_MODIFY_PKT)
			DR_ACTION_INIT_NEXT_STE(DR_STE_TYPE_RX);

		dr_ste_rx_set_flow_tag(last_ste, flow_tag);
	}
}

/* Pay attention that now we support only DR_ACTION_MAX_STES (== 5) ste's */
static int dr_actions_apply(enum dr_ste_entry_type ste_type,
			    uint8_t *action_type_set,
			    uint8_t *last_ste,
			    uint32_t modify_index,
			    uint16_t modify_actions,
			    uint32_t decap_index,
			    uint16_t decap_actions,
			    uint64_t final_icm_addr,
			    uint32_t flow_tag,
			    uint32_t ctr_id,
			    uint16_t gvmi,
			    uint32_t reformat_id,
			    uint32_t reformat_size,
			    struct dr_action_vlan_info *vlans,
			    uint32_t *new_num_stes)
{
	uint32_t added_stes = 0;

	if (ste_type == DR_STE_TYPE_RX) {
		dr_actions_apply_rx(action_type_set, last_ste, gvmi,
				    flow_tag, ctr_id, modify_index,
				    modify_actions, decap_index, decap_actions,
				    vlans, &added_stes);
	} else {
		dr_actions_apply_tx(action_type_set, last_ste, gvmi, ctr_id,
				    modify_index, modify_actions,
				    reformat_id, reformat_size,
				    vlans, &added_stes);
	}

	last_ste += added_stes * DR_STE_SIZE;
	*new_num_stes += added_stes;

	dr_ste_set_hit_addr(last_ste, final_icm_addr, 1);

	return 0;
}

static inline int dr_action_get_domain(enum mlx5dv_dr_ns_domain domain,
				       enum dr_ste_entry_type ste_type)
{
	if (domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS) {
		return DR_ACTION_DOMAIN_NIC_INGRESS;
	} else if (domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS) {
		return DR_ACTION_DOMAIN_NIC_EGRESS;
	} else if (domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS) {
		if (ste_type == DR_STE_TYPE_RX)
			return DR_ACTION_DOMAIN_FDB_INGRESS;
		else if (ste_type == DR_STE_TYPE_TX)
			return DR_ACTION_DOMAIN_FDB_EGRESS;
	}

	return DR_ACTION_DOMAIN_RESERVED;
}

static inline
int dr_action_validate_and_get_next_state(enum dr_action_domain action_domain,
					  uint32_t action_type,
					  uint32_t *state)
{
	uint32_t cur_state = *state;

	/* Check action state machine is valid */
	*state = next_action_state[action_domain][cur_state][action_type];

	if (*state == DR_ACTION_STATE_ERR)
		return EINVAL;

	return 0;
}

#define HDR_LEN_L2_ONLY 14
#define HDR_LEN_L2_VLAN 18

static int dr_actions_l2_re_write(struct mlx5dv_dr_ns *ns,
				  struct mlx5dv_dr_action *action,
				  void *data, size_t data_sz)
{
	uint8_t *l2_hdr = data;
	uint64_t ops[8] = {};
	uint16_t vlan_type;
	bool vlan;
	int i = 0;
	int err;
	void *inl_data;

	vlan = data_sz == HDR_LEN_L2_ONLY ? false : true;

	// TODO replace memcpy with setter and getters.

	/* dmac_47_16 */
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 0);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_0);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 16);
	inl_data = (uint32_t *)(&ops[i]) + 1;
	memcpy(inl_data, l2_hdr, 4);
	i++;

	/* smac_47_16 */
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 0);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_1);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 16);
	inl_data = (uint32_t *)(&ops[i]) + 1;
	memcpy(inl_data, l2_hdr + 6, 4);
	i++;

	/* dmac_15_0 */
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 16);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_0);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 0);
	inl_data = (uint8_t *)(&ops[i]) + 6;
	memcpy(inl_data, l2_hdr + 4, 2);
	i++;

	/* ethertype + (optional) vlan */
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_2);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 32);
	if (!vlan) {
		inl_data = (uint8_t *)(&ops[i]) + 6;
		memcpy(inl_data, l2_hdr + 12, 2);
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 16);
	} else {
		inl_data = (uint32_t *)(&ops[i]) + 1;
		vlan_type = *(uint16_t *)(l2_hdr + 12) == htons(SVLAN_ETHERTYPE) ?
						htons(STE_SVLAN) : htons(STE_CVLAN);
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 18);
		memcpy(inl_data, &vlan_type, 2);
		memcpy(inl_data + 2, l2_hdr + 16, 2);
	}
	i++;

	/* smac_15_0 */
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 16);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_1);
	DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 0);
	inl_data = (uint8_t *)(&ops[i]) + 6;
	memcpy(inl_data, l2_hdr + 10, 2);
	i++;

	if (vlan) {
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, opcode, DR_ACTION_MDFY_HW_OP_SET);
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_length, 16);
		inl_data = (uint8_t *)(&ops[i]) + 6;
		memcpy(inl_data, l2_hdr + 14, 2);
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_field_code, DR_ACTION_MDFY_HW_FLD_L2_2);
		DR_ACTION_SET(dr_action_struct_hw_set, ops + i, destination_left_shifter, 0);
		i++;
	}

	action->re_write.data = (void *)ops;
	action->re_write.num_of_actions = i;
	action->re_write.chunk->byte_size = sizeof(ops);

	err = dr_postsend_action(ns, action);
	if (err) {
		fprintf(stderr, "Writing encapsulation action to ICM failed\n");
		return err;
	}

	return 0;
}

int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
			     struct dr_matcher_rx_tx *nic_matcher,
			     struct mlx5dv_dr_action *actions[],
			     uint32_t num_actions,
			     uint8_t *ste_arr,
			     uint32_t *new_hw_ste_arr_sz)
{
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	uint8_t action_type_set[DR_ACTION_TYP_RESERVED] = {};
	uint64_t final_icm_addr = nic_ns->default_icm_addr;
	bool rx_rule = nic_ns->ste_type == DR_STE_TYPE_RX;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	uint32_t state = DR_ACTION_STATE_NO_ACTION;
	struct dr_action_vlan_info vlans = {};
	enum dr_action_domain action_domain;
	uint16_t gvmi = ns->info.caps.gvmi;
	uint16_t modify_actions = 0;
	uint32_t modify_index = 0;
	uint16_t decap_actions = 0;
	uint32_t decap_index = 0;
	uint32_t flow_tag = 0;
	uint32_t ctr_id = 0;
	uint32_t reformat_id = 0;
	uint32_t reformat_size = 0;
	uint8_t *last_ste;
	int i;

	action_domain = dr_action_get_domain(ns->domain, nic_ns->ste_type);

	for (i = 0; i < num_actions; i++) {
		struct mlx5dv_dr_action *action;
		int max_actions_type = 1;
		uint32_t action_type;

		action = actions[i];
		action_type = action->action_type;

		switch (action_type) {
		case DR_ACTION_TYP_DROP:
			final_icm_addr = nic_ns->drop_icm_addr;
			break;
		case DR_ACTION_TYP_FT:
			if (action->dest_ft->ns != ns) {
				fprintf(stderr, "Destination table belongs to a different NS\n");
				return EINVAL;
			}
			if (action->dest_ft->level <= matcher->tbl->level) {
				fprintf(stderr, "Destination table level is lower than source table level\n");
				return EINVAL;
			}
			final_icm_addr = rx_rule ?
				action->dest_ft->rx.s_anchor->chunk->icm_addr:
				action->dest_ft->tx.s_anchor->chunk->icm_addr;
			break;
		case DR_ACTION_TYP_QP:
			{
				struct mlx5_qp *mlx5_qp = to_mqp(action->qp);

				if (!mlx5_qp->tir_icm_addr) {
					fprintf(stderr, "%s Unsupported QP for action\n", __func__);
					return EINVAL;
				}
				final_icm_addr = mlx5_qp->tir_icm_addr;
			}
			break;
		case DR_ACTION_TYP_CTR:
			ctr_id = action->ctr.devx_obj->object_id + action->ctr.offeset;
			break;
		case DR_ACTION_TYP_TAG:
			flow_tag = action->flow_tag;
			break;
		case DR_ACTION_TYP_TNL_L2_TO_L2:
			break;
		case DR_ACTION_TYP_TNL_L3_TO_L2:
			if (action->re_write.is_root_level) {
				fprintf(stderr, "Decap L3 action cannot be used on current table\n");
				return EINVAL;
			}
			decap_index = action->re_write.index;
			decap_actions = action->re_write.num_of_actions;
			break;
		case DR_ACTION_TYP_MODIFY_HDR:
			if (action->re_write.is_root_level) {
				fprintf(stderr, "Modify header action cannot be used on current table\n");
				return EINVAL;
			}
			modify_index = action->re_write.index;
			modify_actions = action->re_write.num_of_actions;
			break;
		case DR_ACTION_TYP_L2_TO_TNL_L2:
		case DR_ACTION_TYP_L2_TO_TNL_L3:
			if (action->reformat.is_root_level) {
				fprintf(stderr, "Encap action cannot be used on current table\n");
				return EINVAL;
			}
			reformat_size = action->reformat.reformat_size;
			reformat_id = action->reformat.dvo->object_id;
			break;
		case DR_ACTION_TYP_METER:
			if (action->meter.plcr_ft->ns != ns) {
				fprintf(stderr, "Policer table belongs to a different NS\n");
				return EINVAL;
			}
			if (action->meter.plcr_ft->level <= matcher->tbl->level) {
				fprintf(stderr, "Policer table level is lower than source table level\n");
				return EINVAL;
			}
			final_icm_addr = rx_rule ?
				action->meter.rx_icm_addr : action->meter.tx_icm_addr;
			break;
		case DR_ACTION_TYP_VPORT:
			if (action->vport.ns != ns) {
				fprintf(stderr, "Destination vport belongs to a different NS\n");
				return EINVAL;
			}
			if (rx_rule)
				/* If a rule requires to LB traffic from wire,
				 * matching RX packets will be dropped since
				 * we don't support it.
				 */
				if (action->vport.num == WIRE_PORT)
					final_icm_addr = nic_ns->drop_icm_addr;
				else
					final_icm_addr = action->vport.caps->icm_address_rx;
			else
				final_icm_addr = action->vport.caps->icm_address_tx;
			break;
		case DR_ACTION_TYP_POP_VLAN:
			max_actions_type = MAX_VLANS;
			vlans.count++;
			break;
		case DR_ACTION_TYP_PUSH_VLAN:
			max_actions_type = MAX_VLANS;
			if (vlans.count == MAX_VLANS)
				return ENOTSUP;

			vlans.headers[vlans.count++] = action->push_vlan.vlan_hdr;
			break;
		default:
			return EINVAL;
		}

		/* Check action duplication */
		if (++action_type_set[action_type] > max_actions_type) {
			fprintf(stderr, "Action type %d supports only max %d time(s)\n",
				action_type, max_actions_type);
			return EINVAL;
		}

		/* Check action state machine is valid */
		if (dr_action_validate_and_get_next_state(action_domain,
							  action_type,
							  &state)) {
			fprintf(stderr, "Invalid action sequence provided\n");
			return EINVAL;
		}
	}

	*new_hw_ste_arr_sz = nic_matcher->num_of_builders;
	last_ste = ste_arr + DR_STE_SIZE * (nic_matcher->num_of_builders - 1);

	return dr_actions_apply(nic_ns->ste_type, action_type_set, last_ste,
				modify_index, modify_actions,
				decap_index, decap_actions,
				final_icm_addr, flow_tag, ctr_id, gvmi,
				reformat_id, reformat_size, &vlans,
				new_hw_ste_arr_sz);
}

static struct mlx5dv_dr_action *
dr_create_action_generic(enum dr_action_type action_type)
{
	struct mlx5dv_dr_action *action;

	action = calloc(1, sizeof(struct mlx5dv_dr_action));
	if (!action)
		return NULL;

	action->action_type = action_type;
	atomic_init(&action->refcount, 1);

	return action;
}

struct mlx5dv_dr_action *mlx5dv_dr_create_action_drop(void)
{
	return dr_create_action_generic(DR_ACTION_TYP_DROP);
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_dest_ibv_qp(struct ibv_qp *ibqp)
{
	struct mlx5dv_dr_action *action;

	if (ibqp->qp_type != IBV_QPT_RAW_PACKET) {
		fprintf(stderr, "Unsupported QP for action\n");
		errno = EINVAL;
		return NULL;
	}

	action = dr_create_action_generic(DR_ACTION_TYP_QP);
	if (!action)
		return NULL;

	action->qp = ibqp;

	return action;
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_dest_flow_table(struct mlx5dv_dr_tbl *ft)
{
	struct mlx5dv_dr_action *action;

	if (dr_is_root_table(ft)) {
		fprintf(stderr, "Root table cannot be used as a destination\n");
		errno = EINVAL;
		return NULL;
	}

	action = dr_create_action_generic(DR_ACTION_TYP_FT);
	if (!action)
		return NULL;

	action->dest_ft = ft;
	atomic_fetch_add(&ft->refcount, 1);

	return action;
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_devx_counter(struct mlx5dv_devx_obj *devx_obj,
				     uint32_t offeset)
{
	struct mlx5dv_dr_action *action;

	if (devx_obj->type != MLX5_DEVX_FLOW_COUNTER) {
		errno = EINVAL;
		return NULL;
	}

	action = dr_create_action_generic(DR_ACTION_TYP_CTR);
	if (!action)
		return NULL;

	action->ctr.devx_obj = devx_obj;
	action->ctr.offeset = offeset;

	return action;
}

struct mlx5dv_dr_action *mlx5dv_dr_create_action_tag(uint32_t tag_value)
{
	struct mlx5dv_dr_action *action;

	action = dr_create_action_generic(DR_ACTION_TYP_TAG);
	if (!action)
		return NULL;

	action->flow_tag = tag_value & 0xffffff;

	return action;
}

static int
dr_action_create_reformat_action(struct mlx5dv_dr_ns *ns,
				 size_t data_sz,
				 void *data,
				 struct mlx5dv_dr_action *action)
{
	struct ibv_flow_action *flow_action;
	enum mlx5dv_flow_table_type type;

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_RX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_TX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_FDB;
	else
		return EINVAL;

	flow_action = mlx5dv_create_flow_action_packet_reformat(ns->ctx,
								data_sz,
								data,
								action->action_type,
								type);
	if (!flow_action)
		return errno;

	action->reformat.flow_action = flow_action;
	return 0;
}

static int dr_action_verify_reformat_params(enum mlx5dv_flow_action_packet_reformat_type reformat_type,
					    struct mlx5dv_dr_ns *ns,
					    size_t data_sz,
					    void *data)
{
	uint32_t packet_reformat_type = reformat_type;

	if ((!data && data_sz) || (data && !data_sz) ||
	    reformat_type >
	    MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL) {
		fprintf(stderr, "%s invalid parameter!\n", __func__);
		return EINVAL;
	}

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS)
		return 0;

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS) {
		if (packet_reformat_type != MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2 &&
		    packet_reformat_type != MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2) {
			fprintf(stderr, "%s invalid parameter Ingress domain with encap op (%d)\n",
				__func__, reformat_type);
			return EINVAL;
		}
	} else if (ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS) {
		if (packet_reformat_type != MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL &&
		    packet_reformat_type != MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL) {
			fprintf(stderr, "%s invalid parameter Engress domain with decap op (%d)\n",
				__func__, reformat_type);
			return EINVAL;
		}
	} else {
		fprintf(stderr, "%s Unknown NS type (%d)\n",
			__func__, ns->domain);
		return EINVAL;
	}

	return 0;
}
#define ACTION_CACHE_LINE_SIZE 64

static int
dr_action_create_nr_reformat_action(struct mlx5dv_dr_ns *ns,
				    size_t data_sz, void *data,
				    struct mlx5dv_dr_action *action)
{
	struct mlx5dv_devx_obj *obj;

	switch (action->action_type) {
	case DR_ACTION_TYP_L2_TO_TNL_L2:
	case DR_ACTION_TYP_L2_TO_TNL_L3:
	{
		enum reformat_type rt;

		if (action->action_type == DR_ACTION_TYP_L2_TO_TNL_L2)
			rt = MLX5_REFORMAT_TYPE_L2_TO_L2_TUNNEL;
		else
			rt = MLX5_REFORMAT_TYPE_L2_TO_L3_TUNNEL;

		obj = dr_devx_create_reformat_ctx(ns->ctx, rt, data_sz, data);
		if (!obj)
			return errno;

		action->reformat.dvo = obj;
		action->reformat.reformat_size = data_sz;
		return 0;
	}
	case DR_ACTION_TYP_TNL_L2_TO_L2:
	{
		return 0;
	}
	case DR_ACTION_TYP_TNL_L3_TO_L2:
	{
		int err;

		/* only Ethernet frame is supported, either with VLAN (18)
		 * or without (14)
		*/
		if (data_sz != HDR_LEN_L2_ONLY && data_sz != HDR_LEN_L2_VLAN)
			return EINVAL;

		action->re_write.chunk = dr_icm_alloc_mem(ns->action_icm_pool,
							  DR_CHUNK_SIZE_8);
		if (!action->re_write.chunk)
			return ENOMEM;

		action->re_write.index = (action->re_write.chunk->icm_addr -
					  ns->info.caps.hdr_modify_icm_addr) /
					  ACTION_CACHE_LINE_SIZE;

		err = dr_actions_l2_re_write(ns, action, data, data_sz);
		if (err) {
			dr_icm_free_mem(action->re_write.chunk);
			return err;
		}
		return 0;
	}
	default:
		fprintf(stderr, "reformat type not supported %d\n",
			action->action_type);
		return EINVAL;
	}
}

struct mlx5dv_dr_action *mlx5dv_dr_create_action_pop_vlan(void)
{
	return dr_create_action_generic(DR_ACTION_TYP_POP_VLAN);
}

struct mlx5dv_dr_action *mlx5dv_dr_create_action_push_vlan(uint32_t vlan_hdr)
{
	uint32_t vlan_hdr_h = be32toh(vlan_hdr);
	uint16_t ethertype = vlan_hdr_h >> 16;
	struct mlx5dv_dr_action *action;

	if (ethertype != SVLAN_ETHERTYPE && ethertype != CVLAN_ETHERTYPE) {
		fprintf(stderr, "Invalid vlan ethertype\n");
		errno = EINVAL;
		return NULL;
	}

	action = dr_create_action_generic(DR_ACTION_TYP_PUSH_VLAN);
	if (!action)
		return NULL;

	action->push_vlan.vlan_hdr = vlan_hdr_h;
	return action;
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_packet_reformat(struct mlx5dv_dr_ns *ns,
					enum mlx5dv_dr_action_flags action_flags,
					enum mlx5dv_flow_action_packet_reformat_type reformat_type,
					size_t data_sz,
					void *data)
{
	struct mlx5dv_dr_action *action;
	int ret;

	if (action_flags >= MLX5DV_DR_ACTION_FLAGS_RESERVED) {
		errno = EINVAL;
		return NULL;
	}

	if (!ns->info.supp_sw_steering &&
	    !(action_flags & MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)) {
		fprintf(stderr, "Only root actions are supported on ns\n");
		errno = EOPNOTSUPP;
		return NULL;
	}

	/* general checks */
	ret = dr_action_verify_reformat_params(reformat_type, ns, data_sz, data);
	if (ret) {
		errno = ret;
		return NULL;
	}

	action = dr_create_action_generic(reformat_type);
	if (!action)
		return NULL;

	action->reformat.ns = ns;

	/* create the action according to the tbl type */
	if (action_flags & MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL) {
		action->reformat.is_root_level = true;
		ret = dr_action_create_reformat_action(ns,
						       data_sz,
						       data,
						       action);
	} else {
		action->reformat.is_root_level = false;
		ret = dr_action_create_nr_reformat_action(ns,
							  data_sz,
							  data,
							  action);
	}

	if (ret) {
		fprintf(stderr, "%s Failed creating reformat action %d\n",
			__func__, ret);
		goto free_action;
	}

	return action;

free_action:
	errno = ret;
	free(action);
	return NULL;
}

static struct dr_action_modify_field_conv *
dr_action_modify_get_hw_info(uint16_t sw_field)
{
	struct dr_action_modify_field_conv *hw_action_info;

	if (sw_field >= ARRAY_SIZE(dr_action_conv_arr))
		goto not_found;

	hw_action_info = &dr_action_conv_arr[sw_field];
	if (hw_action_info->end - hw_action_info->start == 0)
		goto not_found;

	return hw_action_info;

not_found:
	errno = EINVAL;
	return NULL;
}

static int
dr_action_modify_sw_to_hw(uint64_t *sw_action, uint64_t *hw_action,
			  struct dr_action_modify_field_conv **ret_hw_info)
{
	struct dr_action_modify_field_conv *hw_action_info;
	uint8_t offset, length, max_length, action;
	uint16_t sw_field;
	uint8_t hw_opcode;
	uint32_t data;

	/* Get SW modify action data */
	action = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, action_type);
	sw_field = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, field);
	offset = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, offset);
	length = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, length);
	data = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, data);

	/* Convert SW data to HW modify action format */
	hw_action_info = dr_action_modify_get_hw_info(sw_field);
	if (!hw_action_info) {
		fprintf(stderr, "Modify action invalid field given\n");
		return EINVAL;
	}

	max_length = hw_action_info->end - hw_action_info->start + 1;

	switch (action) {
	case MLX5_ACTION_TYPE_SET:
		hw_opcode = DR_ACTION_MDFY_HW_OP_SET;
		/* The default length is 32 if not set */
		if (!length)
			length = 32;

		if (length + offset > max_length) {
			fprintf(stderr, "Modify action length + offset exceeds limit\n");
			return EINVAL;
		}
		break;

	case MLX5_ACTION_TYPE_ADD:
		hw_opcode = DR_ACTION_MDFY_HW_OP_ADD;
		offset = 0;
		length = max_length;
		break;

	default:
		fprintf(stderr, "Unsupported action_type for modify action\n");
		return EINVAL;
	}

	DR_ACTION_SET(dr_action_struct_hw_set, hw_action,
		      opcode, hw_opcode);

	DR_ACTION_SET(dr_action_struct_hw_set, hw_action,
		      destination_field_code, hw_action_info->hw_field);

	DR_ACTION_SET(dr_action_struct_hw_set, hw_action,
		      destination_left_shifter, hw_action_info->start + offset);

	DR_ACTION_SET(dr_action_struct_hw_set, hw_action,
		      destination_length, length == 32 ? 0 : length);

	DR_ACTION_SET(dr_action_struct_hw_set, hw_action, inline_data, data);

	*ret_hw_info = hw_action_info;

	return 0;
}

static int
dr_action_modify_check_field_limitation(enum mlx5dv_dr_ns_domain domain,
					const uint64_t *sw_action) {

	uint16_t sw_field;
	uint8_t action;

	sw_field = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, field);
	action = DR_ACTION_GET(mlx5_ifc_set_action_in, sw_action, action_type);

	/* Check if SW field is supported in current domain (RX/TX) */
	if (action == MLX5_ACTION_TYPE_SET) {
		if (sw_field == MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGA) {
			if (domain != MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS) {
				fprintf(stderr, "Unsupported field %d for RX/FDB set action\n",
					sw_field);
				return EINVAL;
			}
		}

		if (sw_field == MLX5_ACTION_IN_FIELD_OUT_MEGADATA_REGB) {
			if (domain != MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS) {
				fprintf(stderr, "Unsupported field %d for TX/FDB set action\n",
					sw_field);
				return EINVAL;
			}
		}
	} else if (action == MLX5_ACTION_TYPE_ADD) {
		if (sw_field != MLX5_ACTION_IN_FIELD_OUT_IP_TTL &&
		    sw_field != MLX5_ACTION_IN_FIELD_OUT_IPV6_HOPLIMIT &&
		    sw_field != MLX5_ACTION_IN_FIELD_OUT_TCP_SEQ_NUM &&
		    sw_field != MLX5_ACTION_IN_FIELD_OUT_TCP_ACK_NUM) {
			fprintf(stderr, "Unsupported field %d for add action\n",
				sw_field);
			return EINVAL;
		}
	} else {
		fprintf(stderr, "Unsupported action %d modify action\n", action);
		return EINVAL;
	}

	return 0;
}

static int dr_actions_convert_modify_header(enum mlx5dv_dr_ns_domain domain,
					    uint32_t max_hw_actions,
					    uint32_t num_sw_actions,
					    uint64_t sw_actions[],
					    uint64_t hw_actions[],
					    uint32_t *num_hw_actions)
{
	struct dr_action_modify_field_conv *hw_action_info;
	uint16_t hw_field = DR_ACTION_MDFY_HW_FLD_RESERVED;
	uint32_t l3_type = DR_ACTION_MDFY_HW_HDR_L3_NONE;
	uint32_t l4_type = DR_ACTION_MDFY_HW_HDR_L4_NONE;
	int ret, i, hw_idx = 0;
	uint64_t *sw_action;
	uint64_t hw_action;

	for (i = 0; i < num_sw_actions; i++) {
		sw_action = &sw_actions[i];

		ret = dr_action_modify_check_field_limitation(domain, sw_action);
		if (ret)
			return ret;

		/* Convert SW action to HW action */
		ret = dr_action_modify_sw_to_hw(sw_action,
						&hw_action,
						&hw_action_info);
		if (ret)
			return ret;

		/* Due to a HW limitation we cannot modify 2 different L3 type */
		if (l3_type && hw_action_info->l3_type &&
		    (hw_action_info->l3_type != l3_type)) {
			fprintf(stderr, "Action list can't support 2 different L3 types\n");
			return EINVAL;
		}
		if (hw_action_info->l3_type)
			l3_type = hw_action_info->l3_type;

		/* Due to a HW limitation we cannot modify 2 different L4 type */
		if (l4_type && hw_action_info->l4_type &&
		    (hw_action_info->l4_type != l4_type)) {
			fprintf(stderr, "Action list can't support 2 different L4 types\n");
			return EINVAL;
		}
		if (hw_action_info->l4_type)
			l4_type = hw_action_info->l4_type;

		/* HW reads and executes two actions at once this means we
		 * need to create a gap if two actions access the same field */
		if ((hw_idx % 2) && (hw_field == hw_action_info->hw_field)) {
			/* Check if after gap insertion the total number of HW
			 * modify actions doesn't exceeds the limit */
			hw_idx++;
			if ((num_sw_actions + hw_idx - i) >= max_hw_actions) {
				fprintf(stderr, "Modify header action number exceeds HW limit\n");
				return EINVAL;
			}
		}
		hw_field = hw_action_info->hw_field;

		hw_actions[hw_idx] = hw_action;
		hw_idx++;
	}

	*num_hw_actions = hw_idx;

	return 0;
}

static int
dr_action_create_modify_action(struct mlx5dv_dr_ns *ns,
			       size_t actions_sz,
			       uint64_t actions[],
			       struct mlx5dv_dr_action *action)
{
	struct ibv_flow_action *flow_action;
	enum mlx5dv_flow_table_type type;

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_RX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_TX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_FDB;
	else
		return EINVAL;

	flow_action = mlx5dv_create_flow_action_modify_header(ns->ctx,
							      actions_sz,
							      actions,
							      type);
	if (!flow_action)
		return errno;

	action->re_write.flow_action = flow_action;
	return 0;
}

static int dr_action_create_nr_modify_header(struct mlx5dv_dr_ns *ns,
					     size_t actions_sz,
					     uint64_t actions[],
					     struct mlx5dv_dr_action *action)
{
	struct dr_icm_chunk *chunk;
	uint32_t max_hw_actions;
	uint32_t num_hw_actions;
	uint32_t num_sw_actions;
	uint64_t *hw_actions;
	int ret;

	num_sw_actions = actions_sz / DR_MODIFY_ACTION_SIZE;
	max_hw_actions = dr_icm_pool_chunk_size_to_entries(DR_CHUNK_SIZE_8);

	if (num_sw_actions > max_hw_actions) {
		fprintf(stderr, "Max number of actions %d exceeds limit %d\n",
			num_sw_actions, max_hw_actions);
		errno = EINVAL;
		return errno;
	}

	// TODO use different buckets not just 8 like FW !
	chunk = dr_icm_alloc_mem(ns->action_icm_pool, DR_CHUNK_SIZE_8);
	if (!chunk)
		return errno;

	hw_actions = calloc(1, max_hw_actions * DR_MODIFY_ACTION_SIZE);
	if (!hw_actions) {
		ret = ENOMEM;
		goto free_chunk;
	}

	ret = dr_actions_convert_modify_header(ns->domain,
					       max_hw_actions,
					       num_sw_actions,
					       actions,
					       hw_actions,
					       &num_hw_actions);
	if (ret)
		goto free_hw_actions;

	action->re_write.chunk = chunk;
	action->re_write.data = (uint8_t *)hw_actions;
	action->re_write.num_of_actions = num_hw_actions;
	action->re_write.index = (chunk->icm_addr -
				  ns->info.caps.hdr_modify_icm_addr) /
				  ACTION_CACHE_LINE_SIZE;

	ret = dr_postsend_action(ns, action);
	if (ret)
		goto free_hw_actions;

	return 0;

free_hw_actions:
	free(hw_actions);
free_chunk:
	dr_icm_free_mem(chunk);
	return ret;
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_modify_header(struct mlx5dv_dr_ns *ns,
				      enum mlx5dv_dr_action_flags action_flags,
				      size_t actions_sz,
				      uint64_t actions[])
{
	struct mlx5dv_dr_action *action;
	int ret = 0;

	if (action_flags >= MLX5DV_DR_ACTION_FLAGS_RESERVED) {
		errno = EINVAL;
		return NULL;
	}

	if (actions_sz % DR_MODIFY_ACTION_SIZE) {
		fprintf(stderr, "Invalid modify actions size provided\n");
		errno = EINVAL;
		return NULL;
	}

	if (!ns->info.supp_sw_steering &&
	    !(action_flags & MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)) {
		fprintf(stderr, "Only root actions are supported on ns\n");
		errno = EOPNOTSUPP;
		return NULL;
	}

	action = dr_create_action_generic(DR_ACTION_TYP_MODIFY_HDR);
	if (!action)
		return NULL;

	action->re_write.ns = ns;

	/* create the action according to the tbl type */
	if (action_flags & MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL) {
		action->re_write.is_root_level = true;
		ret = dr_action_create_modify_action(ns,
						     actions_sz,
						     actions,
						     action);
	} else {
		action->re_write.is_root_level = false;
		ret = dr_action_create_nr_modify_header(ns,
							actions_sz,
							actions,
							action);
	}

	if (ret) {
		fprintf(stderr, "%s Failed creating modify header action %d\n",
			__func__, ret);
		goto free_action;
	}

	return action;

free_action:
	errno = ret;
	free(action);
	return NULL;
}

int mlx5dv_dr_modify_action_flow_meter(struct mlx5dv_dr_action *action,
				       struct mlx5dv_dr_flow_meter_init_attr *attr,
				       uint64_t modify_field_select)
{
	int ret;

	if (attr->flow_meter_parameter_sz >
	    sizeof(struct mlx5_ifc_create_flow_meter_in_bits) / 8) {
		errno = EINVAL;
		return errno;
	}

	if (action->action_type != DR_ACTION_TYP_METER) {
		errno = EINVAL;
		return errno;
	}

	ret = dr_devx_modify_meter(action->meter.devx_obj,
				  attr->flow_meter_parameter,
				  attr->flow_meter_parameter_sz,
				  modify_field_select,
				  attr->active);
	return ret;
}

struct mlx5dv_dr_action *
mlx5dv_dr_create_action_flow_meter(struct mlx5dv_dr_flow_meter_init_attr *attr)
{
	struct mlx5dv_dr_ns *ns = attr->policer_ft->ns;
	uint64_t rx_icm_addr, tx_icm_addr;
	struct mlx5dv_devx_obj *devx_obj;
	struct mlx5dv_dr_action *action;
	int ret;

	if (!ns->info.supp_sw_steering) {
		fprintf(stderr, "Action is not supported on current ns\n");
		errno = EOPNOTSUPP;
		return NULL;
	}

	if (dr_is_root_table(attr->policer_ft)) {
		errno = EINVAL;
		return NULL;
	}

	if (attr->flow_meter_parameter_sz >
	    sizeof(struct mlx5_ifc_create_flow_meter_in_bits) / 8) {
		errno = EINVAL;
		return NULL;
	}

	devx_obj = dr_devx_create_meter(ns->ctx,
					attr->active,
					attr->reg_c_index,
					attr->policer_ft->table_type,
					attr->policer_ft->devx_obj->object_id,
					attr->flow_meter_parameter,
					attr->flow_meter_parameter_sz);
	if (!devx_obj)
		return NULL;

	ret = dr_devx_query_meter(devx_obj, &rx_icm_addr, &tx_icm_addr);
	if (ret)
		goto destroy_obj;

	action = dr_create_action_generic(DR_ACTION_TYP_METER);
	if (!action)
		goto destroy_obj;

	action->meter.devx_obj = devx_obj;
	action->meter.plcr_ft = attr->policer_ft;
	action->meter.rx_icm_addr = rx_icm_addr;
	action->meter.tx_icm_addr = tx_icm_addr;

	atomic_fetch_add(&attr->policer_ft->refcount, 1);

	return action;

destroy_obj:
	dr_devx_destroy_meter(devx_obj);
	return NULL;
}

struct mlx5dv_dr_action
*mlx5dv_dr_create_action_dest_vport(struct mlx5dv_dr_ns *ns, uint32_t vport)
{
	struct mlx5dv_dr_action *action;
	struct dr_devx_vport_cap *vport_cap;

	if (!ns->info.supp_sw_steering ||
	    ns->domain != MLX5DV_DR_NS_DOMAIN_FDB_BYPASS) {
		fprintf(stderr, "Domain doesn't support vport actions\n");
		errno = EOPNOTSUPP;
		return NULL;
	}

	vport_cap = dr_get_vport_cap(&ns->info.caps, vport);
	if (!vport_cap)
		return NULL;

	action = dr_create_action_generic(DR_ACTION_TYP_VPORT);
	if (!action)
		return NULL;

	action->vport.ns = ns;
	action->vport.num = vport;
	action->vport.caps = vport_cap;

	return action;
}

int mlx5dv_dr_destroy_action(struct mlx5dv_dr_action *action)
{
	if (atomic_fetch_sub(&action->refcount, 1) > 1)
		return EBUSY;

	switch (action->action_type) {
	case DR_ACTION_TYP_FT:
		atomic_fetch_sub(&action->dest_ft->refcount, 1);
		break;
	case DR_ACTION_TYP_TNL_L2_TO_L2:
		if (action->reformat.is_root_level)
			if (action->reformat.flow_action)
				mlx5_destroy_flow_action(action->reformat.flow_action);
		break;
	case DR_ACTION_TYP_TNL_L3_TO_L2:
		if (action->reformat.is_root_level) {
			if (action->reformat.flow_action)
				mlx5_destroy_flow_action(action->reformat.flow_action);
		} else {
			dr_icm_free_mem(action->re_write.chunk);
		}
		break;
	case DR_ACTION_TYP_L2_TO_TNL_L2:
	case DR_ACTION_TYP_L2_TO_TNL_L3:
		if (action->reformat.is_root_level) {
			if (action->reformat.flow_action)
				mlx5_destroy_flow_action(action->reformat.flow_action);
		} else {
			dr_devx_destroy_reformat_ctx(action->reformat.dvo);
		}
		break;
	case DR_ACTION_TYP_MODIFY_HDR:
		if (action->re_write.is_root_level) {
			if (action->re_write.flow_action)
				mlx5_destroy_flow_action(action->re_write.flow_action);
		} else {
			dr_icm_free_mem(action->re_write.chunk);
			free(action->re_write.data);
		}
		break;
	case DR_ACTION_TYP_METER:
		dr_devx_destroy_meter(action->meter.devx_obj);
		atomic_fetch_sub(&action->meter.plcr_ft->refcount, 1);
		break;
	default:
		break;
	}

	// TODO keep ref on counter/qp/tir?

	free(action);
	return 0;
}

