/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include "mlx5dv_dr.h"

#define DR_MASK_IPV4_ETHERTYPE        0x0800
#define DR_MASK_IPV6_ETHERTYPE        0x86DD
#define DR_MASK_IP_VERSION_IPV4       0x4
#define DR_MASK_IP_VERSION_IPV6       0x6

#define DR_MASK_IS_SMAC_SET(_spec) (_spec.smac_47_16 || _spec.smac_15_0)
#define DR_MASK_IS_DMAC_SET(_spec) (_spec.dmac_47_16 || _spec.dmac_15_0)

#define DR_MASK_IS_IPV6_SRC_ADDR_SET(_spec) (_spec.src_ip_127_96 \
	|| _spec.src_ip_95_64 || _spec.src_ip_63_32 || _spec.src_ip_31_0)

#define DR_MASK_IS_IPV6_DST_ADDR_SET(_spec) (_spec.dst_ip_127_96 \
	|| _spec.dst_ip_95_64 || _spec.dst_ip_63_32 || _spec.dst_ip_31_0)

#define DR_MASK_IS_L3_BASE_SET(_spec) (_spec.ip_protocol \
	|| _spec.frag  || _spec.tcp_flags || _spec.ip_ecn || _spec.ip_dscp)

#define DR_MASK_IS_TCP_UDP_BASE_SET(_spec) (_spec.tcp_sport \
	|| _spec.tcp_dport || _spec.udp_sport || _spec.udp_dport)

#define DR_MASK_IS_IPV4_5_TUPLE_SET(_spec) ( \
	DR_MASK_IS_L3_BASE_SET(_spec) || \
	DR_MASK_IS_TCP_UDP_BASE_SET(_spec) || \
	_spec.dst_ip_31_0 || _spec.src_ip_31_0)

#define DR_MASK_IS_ETH_L2_TNL_SET(_misc) (!!_misc.vxlan_vni)

#define DR_MASK_IS_L2_DST(_spec, _misc, _inner_outer) (_spec.first_vid \
	|| _spec.first_cfi || _spec.first_prio || _spec.cvlan_tag \
	|| _spec.svlan_tag || DR_MASK_IS_DMAC_SET(_spec) || _spec.ethertype \
	|| _spec.ip_version ||  _misc._inner_outer##_second_vid \
	|| _misc._inner_outer##_second_cfi \
	|| _misc._inner_outer##_second_prio \
	|| _misc._inner_outer##_second_cvlan_tag \
	|| _misc._inner_outer##_second_svlan_tag)

#define DR_MASK_IS_TTL_SET(_spec) (!!_spec.ip_ttl_hoplimit)

#define DR_MASK_IS_ETH_L4_SET(_spec, _misc, _inner_outer) ( \
	DR_MASK_IS_L3_BASE_SET(_spec) || \
	DR_MASK_IS_TCP_UDP_BASE_SET(_spec) || \
	DR_MASK_IS_TTL_SET(_spec) || _misc._inner_outer##_ipv6_flow_label)

#define DR_MASK_IS_ETH_L4_MISC_SET(_misc3, _inner_outer) ( \
	!!_misc3._inner_outer##_tcp_seq_num \
	|| !!_misc3._inner_outer##_tcp_ack_num)

#define DR_MASK_IS_FIRST_MPLS_SET(_misc2, _inner_outer) (\
	!!_misc2._inner_outer##_first_mpls_label \
	|| !!_misc2._inner_outer##_first_mpls_exp \
	|| !!_misc2._inner_outer##_first_mpls_s_bos \
	|| !!_misc2._inner_outer##_first_mpls_ttl)

#define DR_MASK_IS_GRE_SET(_misc) (_misc.gre_key_h || _misc.gre_key_l \
	|| _misc.gre_protocol || _misc.gre_c_present \
	|| _misc.gre_k_present || _misc.gre_s_present )

#define DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, gre_udp) (\
	!!_misc2.outer_first_mpls_over_##gre_udp##_label \
	|| !!_misc2.outer_first_mpls_over_##gre_udp##_exp \
	|| !!_misc2.outer_first_mpls_over_##gre_udp##_s_bos \
	|| !!_misc2.outer_first_mpls_over_##gre_udp##_ttl)

#define DR_MASK_IS_FLEX_PARSER_0_SET(_misc2) (\
	DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, gre)\
	|| DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, udp))

#define DR_MASK_IS_FLEX_PARSER_TNL_SET(_misc3) ( \
	_misc3.outer_vxlan_gpe_vni \
	|| _misc3.outer_vxlan_gpe_next_protocol \
	|| _misc3.outer_vxlan_gpe_flags)

#define DR_MASK_IS_FLEX_PARSER_ICMPV6_SET(_misc3) ( \
	_misc3.icmpv6_type \
	|| _misc3.icmpv6_code \
	|| _misc3.icmpv6_header_data)

#define DR_MASK_IS_WQE_METADATA_SET(_misc2)  (!!_misc2.metadata_reg_a)

#define DR_MASK_IS_REG_C_0_3_SET(_misc2) ( \
	!!_misc2.metadata_reg_c_0 ||\
	!!_misc2.metadata_reg_c_1 ||\
	!!_misc2.metadata_reg_c_2 ||\
	!!_misc2.metadata_reg_c_3)

#define DR_MASK_IS_REG_C_4_7_SET(_misc2) ( \
	!!_misc2.metadata_reg_c_4 ||\
	!!_misc2.metadata_reg_c_5 ||\
	!!_misc2.metadata_reg_c_6 ||\
	!!_misc2.metadata_reg_c_7)

#define DR_MASK_IS_GVMI_OR_QPN_SET(_domain, _misc) ( \
	(!!_misc.source_sqn || !!_misc.source_port) &&\
	(_domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||\
	 _domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS))

static inline bool dr_matcher_supp_flex_parser_vxlan_gpe(struct mlx5dv_dr_ns *ns)
{
	return ns->info.caps.flex_protocols &
	       MLX5_FLEX_PARSER_VXLAN_GPE_ENABLED;
}

static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
				       struct dr_matcher_rx_tx *nic_matcher)
{
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct dr_ste_build *sb = nic_matcher->ste_builder;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_match_param mask = {};
	struct dr_match_misc3 *misc3;
	int sb_idx = 0;
	bool inner, rx;
	uint8_t ipv;
	int ret, i;

	if (nic_ns->ste_type == DR_STE_TYPE_RX)
		rx = true;
	else if (nic_ns->ste_type == DR_STE_TYPE_TX)
		rx = false;
	else
		return EINVAL;

	/* Create a temporary mask to track and clear used mask fields */
	if (matcher->match_criteria & DR_MATCHER_CRITERIA_OUTER)
		mask.outer = matcher->mask.outer;

	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC)
		mask.misc = matcher->mask.misc;

	if (matcher->match_criteria & DR_MATCHER_CRITERIA_INNER)
		mask.inner = matcher->mask.inner;

	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC2)
		mask.misc2 = matcher->mask.misc2;

	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC3)
		mask.misc3 = matcher->mask.misc3;

	ret = dr_ste_build_pre_check(matcher->match_criteria,
				     &matcher->mask,
				     NULL);
	if (ret)
		return ret;

	// TODO for now it looks like both inner and outer can be combined to
	// one loop, first outer then inner.

	/* Outer */
	if (matcher->match_criteria & (DR_MATCHER_CRITERIA_OUTER |
				       DR_MATCHER_CRITERIA_MISC |
				       DR_MATCHER_CRITERIA_MISC2 |
				       DR_MATCHER_CRITERIA_MISC3)) {
		inner = false;
		ipv = mask.outer.ip_version;

		if (DR_MASK_IS_WQE_METADATA_SET(mask.misc2)) {
			if (dr_ste_build_general_purpose(&sb[sb_idx++],
							 &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_REG_C_0_3_SET(mask.misc2)) {
			if (dr_ste_build_register_0(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_REG_C_4_7_SET(mask.misc2)) {
			if (dr_ste_build_register_1(&sb[sb_idx++],
						     &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_GVMI_OR_QPN_SET(ns->domain, mask.misc)) {
			if (dr_ste_build_src_gvmi_qpn(&sb[sb_idx++],
						      &mask, &ns->info.caps,
						      inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_SMAC_SET(mask.outer) &&
		    DR_MASK_IS_DMAC_SET(mask.outer)) {
			if (dr_ste_build_eth_l2_src_des(&sb[sb_idx++],
							&mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_SMAC_SET(mask.outer)) {
			if (dr_ste_build_eth_l2_src(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_L2_DST(mask.outer, mask.misc, outer)) {
			if (dr_ste_build_eth_l2_dst(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (ipv == 4) {
			if (DR_MASK_IS_IPV4_5_TUPLE_SET(mask.outer)) {
				if (dr_ste_build_eth_l3_ipv4_5_tuple(&sb[sb_idx++],
								     &mask, inner, rx))
					return EINVAL;
			}

			if (DR_MASK_IS_TTL_SET(mask.outer)) {
				if (dr_ste_build_eth_l3_ipv4_misc(&sb[sb_idx++],
								  &mask, inner, rx))
					return EINVAL;
			}
		} else if (ipv == 6) {
			if (DR_MASK_IS_IPV6_DST_ADDR_SET(mask.outer)) {
				if (dr_ste_build_eth_l3_ipv6_dst(&sb[sb_idx++],
								 &mask, inner, rx))
					return EINVAL;
			}

			if (DR_MASK_IS_IPV6_SRC_ADDR_SET(mask.outer)) {
				if (dr_ste_build_eth_l3_ipv6_src(&sb[sb_idx++],
								 &mask, inner, rx))

					return EINVAL;
			}

			if (DR_MASK_IS_ETH_L4_SET(mask.outer, mask.misc, outer)) {
				if (dr_ste_build_ipv6_l3_l4(&sb[sb_idx++],
							    &mask, inner, rx))
					return EINVAL;
			}
		}

		if (DR_MASK_IS_FLEX_PARSER_TNL_SET(mask.misc3) &&
		    dr_matcher_supp_flex_parser_vxlan_gpe(ns)) {
			if (dr_ste_build_flex_parser_tnl(&sb[sb_idx++],
							 &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_ETH_L4_MISC_SET(mask.misc3, outer)) {
			if (dr_ste_build_eth_l4_misc(&sb[sb_idx++],
						     &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_FIRST_MPLS_SET(mask.misc2, outer)) {
			if (dr_ste_build_mpls(&sb[sb_idx++],
					      &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_FLEX_PARSER_0_SET(mask.misc2)) {
			if (dr_ste_build_flex_parser_0(&sb[sb_idx++],
						       &mask, inner, rx))
				return EINVAL;
		}

		misc3 = &mask.misc3;
		if ((DR_MASK_IS_FLEX_PARSER_ICMPV4_SET(misc3) &&
		    dr_matcher_supp_flex_parser_icmp_v4(&ns->info.caps)) ||
		    (DR_MASK_IS_FLEX_PARSER_ICMPV6_SET(mask.misc3) &&
		     dr_matcher_supp_flex_parser_icmp_v6(&ns->info.caps))) {
			if (dr_ste_build_flex_parser_1(&sb[sb_idx++],
						       &mask, &ns->info.caps,
						       inner,
						       rx))
				return EINVAL;
		}
		if (DR_MASK_IS_GRE_SET(mask.misc)) {
			if (dr_ste_build_gre(&sb[sb_idx++],
					     &mask, inner, rx))
				return EINVAL;
		}
	}

	/* Inner */
	if (matcher->match_criteria & (DR_MATCHER_CRITERIA_INNER |
				       DR_MATCHER_CRITERIA_MISC |
				       DR_MATCHER_CRITERIA_MISC2 |
				       DR_MATCHER_CRITERIA_MISC3)) {
		inner = true;
		ipv = mask.inner.ip_version;

		if (DR_MASK_IS_ETH_L2_TNL_SET(mask.misc)) {
			if (dr_ste_build_eth_l2_tnl(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_SMAC_SET(mask.inner) &&
		    DR_MASK_IS_DMAC_SET(mask.inner)) {
			if (dr_ste_build_eth_l2_src_des(&sb[sb_idx++],
							&mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_SMAC_SET(mask.inner)) {
			if (dr_ste_build_eth_l2_src(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_L2_DST(mask.inner, mask.misc, inner)) {
			if (dr_ste_build_eth_l2_dst(&sb[sb_idx++],
						    &mask, inner, rx))
				return EINVAL;
		}

		if (ipv == 4) {
			if (DR_MASK_IS_IPV4_5_TUPLE_SET(mask.inner)) {
				if (dr_ste_build_eth_l3_ipv4_5_tuple(&sb[sb_idx++],
								     &mask, inner, rx))
					return EINVAL;
			}

			if (DR_MASK_IS_TTL_SET(mask.inner)) {
				if (dr_ste_build_eth_l3_ipv4_misc(&sb[sb_idx++],
								  &mask, inner, rx))
					return EINVAL;
			}
		} else if (ipv == 6) {
			if (DR_MASK_IS_IPV6_DST_ADDR_SET(mask.inner)) {
				if (dr_ste_build_eth_l3_ipv6_dst(&sb[sb_idx++],
								 &mask, inner, rx))
					return EINVAL;
			}

			if (DR_MASK_IS_IPV6_SRC_ADDR_SET(mask.inner)) {
				if (dr_ste_build_eth_l3_ipv6_src(&sb[sb_idx++],
								 &mask, inner, rx))
					return EINVAL;
			}

			if (DR_MASK_IS_ETH_L4_SET(mask.inner, mask.misc, inner)) {
				if (dr_ste_build_ipv6_l3_l4(&sb[sb_idx++],
							    &mask, inner, rx))
					return EINVAL;
			}
		}

		if (DR_MASK_IS_ETH_L4_MISC_SET(mask.misc3, inner)) {
			if (dr_ste_build_eth_l4_misc(&sb[sb_idx++],
						     &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_FIRST_MPLS_SET(mask.misc2, inner)) {
			if (dr_ste_build_mpls(&sb[sb_idx++],
					      &mask, inner, rx))
				return EINVAL;
		}

		if (DR_MASK_IS_FLEX_PARSER_0_SET(mask.misc2)) {
			if (dr_ste_build_flex_parser_0(&sb[sb_idx++],
						       &mask, inner, rx))
				return EINVAL;
		}
	}
	/* empty matcher, takes all */
	if (matcher->match_criteria == DR_MATCHER_CRITERIA_EMPTY) {
		if (dr_ste_build_empty_always_hit(&sb[sb_idx++], rx))
			return EINVAL;
	}

	if (sb_idx == 0) {
		fprintf(stderr, "Cannot generate any valid rules from mask\n");
		return EINVAL;
	}

	nic_matcher->num_of_builders = sb_idx;

	/* Check that all mask fields were consumed */
	for (i = 0; i < sizeof(struct dr_match_param); i++) {
		if (((uint8_t *)&mask)[i] != 0) {
			fprintf(stderr, "Mask contains unsupported parameters\n");
			return EINVAL;
		}
	}
	return 0;
}

static int dr_matcher_connect(struct mlx5dv_dr_ns *ns,
			      struct dr_matcher_rx_tx *curr_nic_matcher,
			      struct dr_matcher_rx_tx *next_nic_matcher,
			      struct dr_matcher_rx_tx *prev_nic_matcher)
{
	struct dr_tbl_rx_tx *nic_tbl = curr_nic_matcher->nic_tbl;
	struct dr_ns_rx_tx *nic_ns = nic_tbl->nic_ns;
	struct dr_htbl_connect_info info;
	struct dr_ste_htbl *prev_htbl;
	int ret;

	/* Connect end anchor hash table to next_htbl or to the default address */
	if (next_nic_matcher) {
		info.type = CONNECT_HIT;
		info.hit_next_htbl = next_nic_matcher->s_htbl;
	} else {
		info.type = CONNECT_MISS;
		info.miss_icm_addr = nic_ns->default_icm_addr;
	}
	ret = dr_ste_htbl_init_and_postsend(ns, nic_ns,
					    curr_nic_matcher->e_anchor,
					    &info, info.type == CONNECT_HIT);
	if (ret)
		return ret;

	/* Connect start hash table to end anchor */
	info.type = CONNECT_MISS;
	info.miss_icm_addr = curr_nic_matcher->e_anchor->chunk->icm_addr;
	ret = dr_ste_htbl_init_and_postsend(ns, nic_ns,
					    curr_nic_matcher->s_htbl,
					    &info, false);
	if (ret)
		return ret;

	/* Connect previous hash table to matcher start hash table */
	if (prev_nic_matcher)
		prev_htbl = prev_nic_matcher->e_anchor;
	else
		prev_htbl = nic_tbl->s_anchor;

	info.type = CONNECT_HIT;
	info.hit_next_htbl = curr_nic_matcher->s_htbl;
	ret = dr_ste_htbl_init_and_postsend(ns, nic_ns, prev_htbl,
					    &info, true);
	if (ret)
		return ret;

	/* Update the pointing ste and next hash table */
	// TODO ESW have these pointing ste mess in a func!
	curr_nic_matcher->s_htbl->pointing_ste = prev_htbl->ste_arr;
	prev_htbl->ste_arr[0].next_htbl = curr_nic_matcher->s_htbl;

	if (next_nic_matcher) {
		next_nic_matcher->s_htbl->pointing_ste = curr_nic_matcher->e_anchor->ste_arr;
		curr_nic_matcher->e_anchor->ste_arr[0].next_htbl = next_nic_matcher->s_htbl;
	}

	return 0;
}

static int dr_matcher_add_to_tbl(struct mlx5dv_dr_matcher *matcher)
{
	struct mlx5dv_dr_matcher *next_matcher, *prev_matcher;
	struct mlx5dv_dr_tbl *tbl = matcher->tbl;
	int ret, i, prev_limit, next_limit;
	struct mlx5dv_dr_ns *ns = tbl->ns;

	if (dr_is_root_table(matcher->tbl))
		return 0;

	pthread_mutex_lock(&tbl->mutex);
	pthread_mutex_lock(&matcher->mutex);

	prev_limit = -1;
	for (i = matcher->prio - 1; i >= 0; i--)
		if (tbl->prio[i]) {
			prev_limit = i;
			break;
		}

	next_limit = -1;
	for (i = matcher->prio; i < tbl->max_prio; i++)
		if (tbl->prio[i]) {
			next_limit = i;
			break;
		}

	if (next_limit != -1)
		next_matcher = tbl->prio[next_limit];
	else
		next_matcher = NULL;

	if (prev_limit != -1)
		if (next_matcher)
			/* Matcher located middle in list */
			prev_matcher = list_prev(&tbl->matcher_list,
						 next_matcher,
						 matcher_list);
		else
			/* Matcher located last in list */
			prev_matcher = list_tail(&tbl->matcher_list,
						 struct mlx5dv_dr_matcher,
						 matcher_list);
	else
		/* Matcher located first in list */
		prev_matcher = NULL;

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS) {
		ret = dr_matcher_connect(ns, &matcher->rx,
					 next_matcher ? &next_matcher->rx : NULL,
					 prev_matcher ?	&prev_matcher->rx : NULL);
		if (ret)
			goto out_unlock;
	}

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS) {
		ret = dr_matcher_connect(ns, &matcher->tx,
					 next_matcher ? &next_matcher->tx : NULL,
					 prev_matcher ?	&prev_matcher->tx : NULL);
		if (ret)
			goto out_unlock;
	}

	tbl->prio[matcher->prio] = matcher;

	if (prev_matcher)
		list_add_after(&tbl->matcher_list,
			       &prev_matcher->matcher_list,
			       &matcher->matcher_list);
	else if (next_matcher)
		list_add_before(&tbl->matcher_list,
				&next_matcher->matcher_list,
				&matcher->matcher_list);
	else
		list_add(&tbl->matcher_list, &matcher->matcher_list);

out_unlock:
	pthread_mutex_unlock(&matcher->mutex);
	pthread_mutex_unlock(&tbl->mutex);
	return ret;
}

static void dr_matcher_uninit_nic(struct dr_matcher_rx_tx *nic_matcher)
{
	dr_htbl_put(nic_matcher->s_htbl);
	dr_htbl_put(nic_matcher->e_anchor);
}

static void dr_matcher_uninit_fdb(struct mlx5dv_dr_matcher *matcher)
{
	dr_matcher_uninit_nic(&matcher->rx);
	dr_matcher_uninit_nic(&matcher->tx);
}

static int dr_matcher_uninit_root(struct mlx5dv_dr_matcher *matcher)
{
	return mlx5dv_destroy_flow_matcher(matcher->dv_matcher);
}

static void dr_matcher_uninit(struct mlx5dv_dr_matcher *matcher)
{
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;

	if (dr_is_root_table(matcher->tbl)) {
		dr_matcher_uninit_root(matcher);
		return;
	}

	switch (ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		dr_matcher_uninit_nic(&matcher->rx);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		dr_matcher_uninit_nic(&matcher->tx);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		dr_matcher_uninit_fdb(matcher);
		break;
	default:
		errno = EINVAL;
		break;
	}
}

static int dr_matcher_init_nic(struct mlx5dv_dr_matcher *matcher,
			       struct dr_matcher_rx_tx *nic_matcher)
{
	struct dr_tbl_rx_tx *nic_tbl = nic_matcher->nic_tbl;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	uint32_t tbl_size;
	int ret;

	ret = dr_matcher_set_ste_builders(matcher, nic_matcher);
	if (ret)
		return ret;

	nic_matcher->e_anchor = dr_ste_htbl_alloc(ns->ste_icm_pool,
						  DR_CHUNK_SIZE_1,
						  DR_STE_LU_TYPE_DONT_CARE,
						  0);
	if (!nic_matcher->e_anchor)
		return ret;

	if (matcher->match_criteria == DR_MATCHER_CRITERIA_EMPTY)
		tbl_size = DR_CHUNK_SIZE_1;
	else
		tbl_size = nic_tbl->nic_ns->table_sizes[0];

	nic_matcher->s_htbl = dr_ste_htbl_alloc(ns->ste_icm_pool,
						tbl_size,
						nic_matcher->ste_builder[0].lu_type,
						nic_matcher->ste_builder[0].byte_mask);
	if (!nic_matcher->s_htbl)
		goto free_e_htbl;

	/* make sure the tables exist while empty */
	dr_htbl_get(nic_matcher->s_htbl);
	dr_htbl_get(nic_matcher->e_anchor);

	return 0;

free_e_htbl:
	dr_ste_htbl_free(nic_matcher->e_anchor);
	return errno;
}

static int dr_matcher_init_fdb(struct mlx5dv_dr_matcher *matcher)
{
	int ret;

	ret = dr_matcher_init_nic(matcher, &matcher->rx);
	if (ret)
		return ret;

	ret = dr_matcher_init_nic(matcher, &matcher->tx);
	if (ret)
		goto uninit_nic_rx;

	return 0;

uninit_nic_rx:
	dr_matcher_uninit_nic(&matcher->rx);
	return ret;
}

static int dr_matcher_init_root(struct mlx5dv_dr_matcher *matcher,
				struct mlx5dv_flow_match_parameters *mask)
{
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct mlx5dv_flow_matcher_attr attr = {};
	enum mlx5dv_flow_table_type type;

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_RX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_TX;
	else if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS)
		type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_FDB;
	else
		return EINVAL;

	attr.match_mask = mask;
	attr.priority = matcher->prio;
	attr.type = IBV_FLOW_ATTR_NORMAL;
	attr.match_criteria_enable = matcher->match_criteria;
	attr.ft_type = type;
	attr.comp_mask = MLX5DV_FLOW_MATCHER_MASK_FT_TYPE;

	matcher->dv_matcher = mlx5dv_create_flow_matcher(ns->ctx, &attr);
	if (!matcher->dv_matcher)
		return errno;

	return 0;
}

static int dr_matcher_init(struct mlx5dv_dr_matcher *matcher,
			   struct mlx5dv_flow_match_parameters *mask)
{
	struct mlx5dv_dr_tbl *tbl = matcher->tbl;
	struct mlx5dv_dr_ns *ns = tbl->ns;
	int ret;

	if (dr_is_root_table(matcher->tbl))
		return dr_matcher_init_root(matcher, mask);

	if (matcher->match_criteria >= DR_MATCHER_CRITERIA_MAX) {
		fprintf(stderr, "Invalid match criteria attribute\n");
		errno = EINVAL;
		return errno;
	}

	if (mask && (mask->match_sz > sizeof(struct dr_match_param) ||
	    (mask->match_sz % sizeof(uint32_t)))) {
		fprintf(stderr, "Invalid match size attribute\n");
		errno = EINVAL;
		return errno;
	}

	if (mask) {
		memcpy(&matcher->mask, mask->match_buf, mask->match_sz);
		dr_be32_buffer_to_h32((uint64_t *)&matcher->mask, mask->match_sz);
	}

	switch (ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		matcher->rx.nic_tbl = &tbl->rx;
		ret = dr_matcher_init_nic(matcher, &matcher->rx);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		matcher->tx.nic_tbl = &tbl->tx;
		ret = dr_matcher_init_nic(matcher, &matcher->tx);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		matcher->rx.nic_tbl = &tbl->rx;
		matcher->tx.nic_tbl = &tbl->tx;
		ret = dr_matcher_init_fdb(matcher);
		break;
	default:
		ret = EINVAL;
		errno = ret;
		break;
	}

	return ret;
}

struct mlx5dv_dr_matcher *
mlx5dv_dr_create_matcher(struct mlx5dv_dr_tbl *tbl,
			 uint16_t priority,
			 uint8_t match_criteria_enable,
			 struct mlx5dv_flow_match_parameters *mask)
{
	struct mlx5dv_dr_matcher *matcher;
	int ret;

	if (priority >= tbl->max_prio) {
		errno = EINVAL;
		return NULL;
	}

	matcher = calloc(1, sizeof(*matcher));
	if (!matcher)
		return NULL;

	matcher->tbl = tbl;
	matcher->prio = priority;
	matcher->match_criteria = match_criteria_enable;
	atomic_init(&matcher->refcount, 1);
	pthread_mutex_init(&matcher->mutex, NULL);
	list_node_init(&matcher->matcher_list);

	ret = dr_matcher_init(matcher, mask);
	if (ret)
		goto free_matcher;

	ret = dr_matcher_add_to_tbl(matcher);
	if (ret)
		goto matcher_uninit;

	atomic_fetch_add(&tbl->refcount, 1);

	return matcher;

matcher_uninit:
	dr_matcher_uninit(matcher);
free_matcher:
	pthread_mutex_destroy(&matcher->mutex);
	free(matcher);
	return NULL;
}

static int dr_matcher_disconnect(struct mlx5dv_dr_ns *ns,
				 struct dr_tbl_rx_tx *nic_tbl,
				 struct dr_matcher_rx_tx *next_nic_matcher,
				 struct dr_matcher_rx_tx *prev_nic_matcher)
{
	struct dr_ns_rx_tx *nic_ns = nic_tbl->nic_ns;
	struct dr_htbl_connect_info info;
	struct dr_ste_htbl *prev_anchor;

	if (prev_nic_matcher)
		prev_anchor = prev_nic_matcher->e_anchor;
	else
		prev_anchor = nic_tbl->s_anchor;

	/* Connect previous anchor hash table to next matcher or to the default address */
	if (next_nic_matcher) {
		info.type = CONNECT_HIT;
		info.hit_next_htbl = next_nic_matcher->s_htbl;
		next_nic_matcher->s_htbl->pointing_ste = prev_anchor->ste_arr;
		prev_anchor->ste_arr[0].next_htbl = next_nic_matcher->s_htbl;
	} else {
		info.type = CONNECT_MISS;
		info.miss_icm_addr = nic_ns->default_icm_addr;
		prev_anchor->ste_arr[0].next_htbl = NULL;
	}

	return dr_ste_htbl_init_and_postsend(ns, nic_ns, prev_anchor,
					     &info, true);
}

static int dr_matcher_remove_from_tbl(struct mlx5dv_dr_matcher *matcher)
{
	struct mlx5dv_dr_matcher *prev_matcher, *next_matcher;
	struct mlx5dv_dr_tbl *tbl = matcher->tbl;
	struct mlx5dv_dr_ns *ns = tbl->ns;
	int ret = 0;

	if (dr_is_root_table(matcher->tbl))
		return 0;

	pthread_mutex_lock(&tbl->mutex);
	pthread_mutex_lock(&matcher->mutex);

	prev_matcher = list_prev(&tbl->matcher_list, matcher, matcher_list);
	next_matcher = list_next(&tbl->matcher_list, matcher, matcher_list);

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS) {
		ret = dr_matcher_disconnect(ns, &tbl->rx,
					    next_matcher ? &next_matcher->rx : NULL,
					    prev_matcher ? &prev_matcher->rx : NULL);
		if (ret)
			goto out_unlock;
	}

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS) {
		ret = dr_matcher_disconnect(ns, &tbl->tx,
					    next_matcher ? &next_matcher->tx : NULL,
					    prev_matcher ? &prev_matcher->tx : NULL);
		if (ret)
			goto out_unlock;
	}

	if (tbl->prio[matcher->prio] == matcher) {
		if (next_matcher && next_matcher->prio == matcher->prio)
			tbl->prio[matcher->prio] = next_matcher;
		else
			tbl->prio[matcher->prio] = NULL;
	}

	list_del(&matcher->matcher_list);

out_unlock:
	pthread_mutex_unlock(&matcher->mutex);
	pthread_mutex_unlock(&tbl->mutex);
	return ret;
}

int mlx5dv_dr_destroy_matcher(struct mlx5dv_dr_matcher *matcher)
{
	if (atomic_load(&matcher->refcount) > 1)
		return EBUSY;

	dr_matcher_remove_from_tbl(matcher);
	dr_matcher_uninit(matcher);
	atomic_fetch_sub(&matcher->tbl->refcount, 1);
	pthread_mutex_destroy(&matcher->mutex);
	free(matcher);
	return 0;
}
